from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import override_settings
from arkindex.project.tests import FixtureTestCase
from arkindex_common.enums import DataImportMode
from arkindex.dataimport.models import DataImport
import yaml


class TestImportS3(FixtureTestCase):

    @override_settings(
        AWS_ACCESS_KEY='username',
        AWS_SECRET_KEY='s3kr3t',
        AWS_ENDPOINT='http://nowhere/',
        AWS_REGION='middle-earth',
        ARKINDEX_TASKS_IMAGE='tasks:latest',
        PONOS_RECIPE={
            'env': {
                'SOME_VAR': 'somevalue',
            }
        }
    )
    def test_dest_bucket(self):
        self.assertFalse(DataImport.objects.exists())

        call_command(
            'import_s3',
            source_bucket='mybucket',
            dest_bucket='somewhere',
            corpus=self.corpus,
            bucket_prefix='A',
            max_folder_depth=3,
            nb_chunks=2,
        )

        di = DataImport.objects.get()
        self.assertEqual(di.mode, DataImportMode.Images)
        self.assertEqual(di.corpus, self.corpus)
        self.assertEqual(di.creator, self.superuser)
        self.assertIsNone(di.element)

        recipe = yaml.safe_load(di.workflow.recipe)
        self.assertDictEqual(recipe, {
            'env': {
                'AWS_ACCESS_KEY': 'username',
                'AWS_SECRET_KEY': 's3kr3t',
                'AWS_ENDPOINT': 'http://nowhere/',
                'AWS_REGION': 'middle-earth',
                'SOME_VAR': 'somevalue',
            },
            'tasks': {
                'import-s3': {
                    'image': 'tasks:latest',
                    'command': 'python3 -m arkindex_tasks.import_s3 '
                               '--source-bucket mybucket '
                               '--corpus-id {} '
                               "--bucket-prefix 'A' "
                               '--max-folder-depth 3 '
                               '--nb-chunks 2 '
                               '--dest-bucket somewhere'.format(self.corpus.id),
                },
                'import_chunk_1': {
                    'image': 'tasks:latest',
                    'command': 'python3 -m arkindex_tasks.import_s3.volumes_import /data/import-s3/chunk_1.yml',
                    'parents': ['import-s3'],
                },
                'import_chunk_2': {
                    'image': 'tasks:latest',
                    'command': 'python3 -m arkindex_tasks.import_s3.volumes_import /data/import-s3/chunk_2.yml',
                    'parents': ['import-s3'],
                },
                'thumbnails_chunk_1': {
                    'image': 'tasks:latest',
                    'command': 'python3 -m arkindex_tasks.generate_thumbnails /data/import_chunk_1/elements.json',
                    'parents': ['import_chunk_1'],
                },
                'thumbnails_chunk_2': {
                    'image': 'tasks:latest',
                    'command': 'python3 -m arkindex_tasks.generate_thumbnails /data/import_chunk_2/elements.json',
                    'parents': ['import_chunk_2'],
                }
            }
        })

    @override_settings(
        AWS_ACCESS_KEY='username',
        AWS_SECRET_KEY='s3kr3t',
        AWS_ENDPOINT='http://nowhere/',
        AWS_REGION='middle-earth',
        ARKINDEX_TASKS_IMAGE='tasks:latest',
    )
    def test_iiif_cache_bucket(self):
        self.assertFalse(DataImport.objects.exists())

        call_command(
            'import_s3',
            source_bucket='mybucket',
            image_server=42,
            corpus=self.corpus,
            bucket_prefix='A',
            max_folder_depth=2,
            nb_chunks=1,
            iiif_cache_bucket='melon_cache',
        )

        di = DataImport.objects.get()
        recipe = yaml.safe_load(di.workflow.recipe)
        self.assertDictEqual(recipe.get('tasks'), {
            'import-s3': {
                'image': 'tasks:latest',
                'command': 'python3 -m arkindex_tasks.import_s3 '
                           '--source-bucket mybucket '
                           '--corpus-id {} '
                           "--bucket-prefix 'A' "
                           '--max-folder-depth 2 '
                           '--nb-chunks 1 '
                           '--iiif-cache-bucket melon_cache '
                           '--image-server 42'.format(self.corpus.id),
            },
            'import_chunk_1': {
                'image': 'tasks:latest',
                'command': 'python3 -m arkindex_tasks.import_s3.volumes_import /data/import-s3/chunk_1.yml',
                'parents': ['import-s3'],
            },
            'thumbnails_chunk_1': {
                'image': 'tasks:latest',
                'command': 'python3 -m arkindex_tasks.generate_thumbnails /data/import_chunk_1/elements.json',
                'parents': ['import_chunk_1'],
            },
            'cache_builder_1': {
                'image': 'tasks:latest',
                'command': (
                    'python3 -m arkindex_tasks.import_s3.build_cantaloupe_cache '
                    '/data/import_chunk_1/s3_elements.json'
                ),
                'parents': ['import_chunk_1'],
            }
        })

    def test_xor_args(self):
        with self.assertRaises(CommandError):
            call_command(
                'import_s3',
                source_bucket='mybucket',
                dest_bucket='somewhere',
                image_server='some_server',
                corpus=self.corpus,
                bucket_prefix='A',
                max_folder_depth=3,
                nb_chunks=2,
            )
