from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import DataImportMode
from arkindex.dataimport.models import Worker, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from rest_framework.exceptions import ValidationError


class TestSignals(FixtureAPITestCase):
    """
    Test signals
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get()
        cls.rev_1 = cls.repo.revisions.get()
        cls.rev_2 = cls.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )

        cls.worker_1 = Worker.objects.create(
            repository=cls.repo,
            name='Worker 1',
            slug='worker_1',
            type=MLToolType.Classifier
        )
        cls.version_1 = WorkerVersion.objects.create(
            worker=cls.worker_1,
            revision=cls.rev_1,
            configuration={"test": "test1"}
        )
        cls.version_2 = WorkerVersion.objects.create(
            worker=cls.worker_1,
            revision=cls.rev_2,
            configuration={"test": "test2"}
        )

        cls.dataimport_1 = cls.corpus.imports.create(creator=cls.user, mode=DataImportMode.Images)
        cls.run_1 = cls.dataimport_1.worker_runs.create(
            version=cls.version_1,
            parents=[],
        )
        cls.dataimport_2 = cls.corpus.imports.create(creator=cls.user, mode=DataImportMode.Images)

    def test_worker_run_check_parents_recursive(self):
        run_2 = self.dataimport_1.worker_runs.create(
            version=self.version_2,
            parents=[],
        )
        self.assertListEqual(run_2.parents, [])

        run_2.parents = [str(run_2.id)]
        with self.assertRaises(ValidationError):
            run_2.save()

        run_2.refresh_from_db()
        self.assertListEqual(run_2.parents, [])

    def test_worker_run_check_parents_distinct_dataimport(self):
        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)
        with self.assertRaises(ValidationError):
            self.dataimport_2.worker_runs.create(
                version=self.version_2,
                parents=[str(self.run_1.id)],
            )
        self.dataimport_1.refresh_from_db()

        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)

    def test_worker_run_check_parents_inexsistant_uuid(self):
        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)
        with self.assertRaises(ValidationError):
            self.dataimport_2.worker_runs.create(
                version=self.version_2,
                parents=['12341234-1234-1234-1234-123412341234'],
            )
        self.dataimport_1.refresh_from_db()

        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)

    def test_worker_run_check_parents(self):
        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)
        run_2 = self.dataimport_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )
        self.assertTrue(True)

        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 2)
        self.assertListEqual(list(self.dataimport_1.worker_runs.all()), [
            self.run_1,
            run_2,
        ])
