from django.test import override_settings
from django.urls import reverse
from rest_framework import status
from arkindex.documents.models import Corpus, Element
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User


class TestElementsAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.act_type = cls.corpus.types.get(slug='act')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)

    def test_select_element_not_login(self):
        response = self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id]},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={'selection': False})
    def test_select_element_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id]},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ['Selection is not available on this instance.'])

    def test_select_element_wrong_corpus(self):
        user = User.objects.create_user('nope@nope.fr')
        self.client.force_login(user)
        element = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name='Act xx'
        )
        response = self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [element.id]},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'ids': ['Some element IDs do not exist or you do not have permission to access them.'],
        })

    def test_select_element_wrong_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id, 'ffffffff-ffff-ffff-ffff-ffffffffffff']},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'ids': ['Some element IDs do not exist or you do not have permission to access them.'],
        })

    def test_select_element_bad_uuid(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id, 'oh no']},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'1': ['Must be a valid UUID.']})

    def test_select_element_no_id(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_select_element(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id]},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        results = response.json()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['id'], str(self.vol.id))
        response = self.client.get(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['id'], str(self.vol.id))

        response = self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id, self.page.id]},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        results = response.json()
        self.assertEqual(len(results), 2)
        self.assertCountEqual(
            [result['id'] for result in results],
            [str(self.page.id), str(self.vol.id)]
        )
        response = self.client.get(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 2)
        self.assertCountEqual(
            [result['id'] for result in results],
            [str(self.page.id), str(self.vol.id)]
        )

        response = self.client.get(reverse('api:elements-selection') + '?with_best_classes=true')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 2)
        self.assertCountEqual(
            [result['id'] for result in results],
            [str(self.page.id), str(self.vol.id)]
        )
        self.assertEqual(results[0]['best_classes'], [])
        self.assertEqual(results[1]['best_classes'], [])

    def test_delete_wrong_id(self):
        self.client.force_login(self.user)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id]},
            format='json'
        )
        response = self.client.delete(reverse('api:elements-selection'), data={'id': self.page.id})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @override_settings(ARKINDEX_FEATURES={'selection': False})
    def test_delete_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:elements-selection'), data={'id': self.page.id})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ['Selection is not available on this instance.'])

    def test_delete_selection(self):
        self.client.force_login(self.user)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id, self.page.id]},
            format='json'
        )
        response = self.client.get(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertCountEqual(
            [result['id'] for result in results],
            [str(self.page.id), str(self.vol.id)]
        )

        response = self.client.delete(reverse('api:elements-selection'), data={'id': self.vol.id})
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        response = self.client.get(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['id'], str(self.page.id))

    def test_select_element_two_session(self):
        self.client.force_login(self.user)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id]},
            format='json'
        )
        response = self.client.get(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['id'], str(self.vol.id))

        user = User.objects.create_user('nope@nope.fr')
        self.client.force_login(user)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.page.id]},
            format='json'
        )
        response = self.client.get(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['id'], str(self.page.id))

    def test_delete_all_selection(self):
        self.client.force_login(self.user)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.vol.id, self.page.id]},
            format='json'
        )
        response = self.client.get(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 2)
        self.assertCountEqual(
            [result['id'] for result in results],
            [str(self.page.id), str(self.vol.id)]
        )

        response = self.client.delete(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        response = self.client.get(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 0)

    @override_settings(ARKINDEX_FEATURES={'selection': False})
    def test_list_selection_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:elements-selection'))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ['Selection is not available on this instance.'])
