from unittest.mock import patch, call, MagicMock
from django.test import TestCase, override_settings
from rest_framework.serializers import ValidationError
from arkindex_common.ml_tool import MLTool, MLToolType
from arkindex.project.serializer_fields import DataSourceSlugField
from arkindex.documents.models import DataSource


@override_settings(ML_CLASSIFIERS_DIR='/somewhere')
class TestDataSourceSlugField(TestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.external_source = DataSource.objects.create(
            name='External sauce',
            type=MLToolType.Classifier,
            slug='sauce',
            internal=False,
        )
        cls.internal_source = DataSource.objects.create(
            name='Internal sauce',
            type=MLToolType.Recognizer,
            slug='better_sauce',
            revision='13.37',
            internal=True,
        )

    def test_init(self):
        with self.assertRaises(AssertionError):
            DataSourceSlugField(tool_type='oops')

    def test_to_representation(self):
        field = DataSourceSlugField()
        self.assertEqual(field.to_representation('something'), 'something')
        self.assertEqual(field.to_representation(self.external_source), 'sauce')

    def test_to_internal_value(self):
        field = DataSourceSlugField()
        self.assertEqual(field.to_internal_value('sauce'), self.external_source)
        with self.assertRaises(ValidationError, msg="Source with slug 'mayo' not found"):
            field.to_internal_value('mayo')

    def test_to_internal_value_latest(self):
        """
        Test DataSourceSlugField.to_internal_value picks the latest source by revision
        """
        field = DataSourceSlugField()
        DataSource.objects.create(
            name='Spoiled mayonnaise',
            type=MLToolType.Recognizer,
            slug='mayo',
            revision='42',
            internal=False,
        )
        mayo = DataSource.objects.create(
            name='Edible mayonnaise',
            type=MLToolType.Recognizer,
            slug='mayo',
            revision='43',
            internal=False,
        )
        self.assertEqual(field.to_internal_value('mayo'), mayo)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_to_internal_value_tool_type(self, ml_get_mock):
        tool_mock = MagicMock(spec=MLTool)
        tool_mock.name = 'Green pesto'
        tool_mock.slug = 'pesto'
        tool_mock.version = '10.42'
        tool_mock.type = MLToolType.Recognizer
        ml_get_mock.return_value = tool_mock

        field = DataSourceSlugField(tool_type=MLToolType.Recognizer)
        self.assertFalse(DataSource.objects.filter(slug='pesto').exists())
        pesto_source = field.to_internal_value('pesto')
        self.assertTrue(DataSource.objects.filter(slug='pesto').exists())

        self.assertEqual(pesto_source.type, MLToolType.Recognizer)
        self.assertEqual(pesto_source.slug, 'pesto')
        self.assertEqual(pesto_source.name, 'Green pesto')
        self.assertEqual(pesto_source.revision, '10.42')
        self.assertTrue(pesto_source.internal)

        self.assertEqual(ml_get_mock.call_count, 1)
        self.assertEqual(ml_get_mock.call_args, call('/somewhere', MLToolType.Recognizer, 'pesto'))

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_to_internal_value_tool_type_not_found(self, ml_get_mock):
        ml_get_mock.side_effect = ValueError

        field = DataSourceSlugField(tool_type=MLToolType.Classifier)
        with self.assertRaises(ValidationError, msg="Source with slug 'better_sauce' not found"):
            field.to_internal_value('better_sauce')

        field = DataSourceSlugField(tool_type=MLToolType.Recognizer)
        self.assertEqual(field.to_internal_value('better_sauce'), self.internal_source)

        self.assertEqual(ml_get_mock.call_count, 2)
        self.assertEqual(ml_get_mock.call_args_list, [
            call('/somewhere', MLToolType.Classifier, 'better_sauce'),
            call('/somewhere', MLToolType.Recognizer, 'better_sauce'),
        ])
